@@ -6,11 +6,11 @@
     
 ======================================================================*/
 
-#include <linux/autoconf.h>
+#include "config.h"
+#include "k_compat.h"
 
 #ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_i82365 init_module
 #endif
 
 #include <sys/types.h>
@@ -28,6 +28,7 @@
 #include <linux/sched.h>
 #include <linux/malloc.h>
 #include <linux/pci.h>
+#include <linux/bios32.h>
 #include <linux/ioport.h>
 
 #include "version.h"
@@ -42,7 +43,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "i82365.c 1.70 1995/07/06 19:39:18 (David Hinds)\n";
+"i82365.c 1.95 1996/05/31 05:28:24 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -55,6 +56,9 @@ static int i365_base = 0x3e0;
 /* Specify a socket number to ignore */
 static int ignore = -1;
 
+/* Should we probe at 0x3e2 for an extra controller? */
+static int extra_sockets = 0;
+
 /* Bit map of interrupts to choose from -- 0 means just autoprobe */
 static u_long irq_mask = 0;
 
@@ -64,8 +68,8 @@ static int cs_irq = 0;
 /* Poll status interval -- 0 means default to interrupt */
 static int poll_interval = 0;
 
-/* External clock time, in nanoseconds.  210 ns = 4.77 MHz */
-static int cycle_time = 210;
+/* External clock time, in nanoseconds.  120 ns = 8.33 MHz */
+static int cycle_time = 120;
 
 /* These features may or may not be implemented */
 static int has_dma = 0;
@@ -91,7 +95,7 @@ static int wakeup = 0;
 
 static struct timer_list poll_timer;
 
-static void i365_interrupt(int irq, struct pt_regs *regs);
+static void i365_interrupt IRQ(int irq, void *dev, struct pt_regs *regs);
 
 static int i365_service(u_int sock, u_int cmd, void *arg);
 
@@ -105,16 +109,17 @@ typedef struct socket_info_t {
 } socket_info_t;
 
 /* Controller types */
-#define CHIP(t)		((t) & 0xf0)
-#define ID(t)		((t) & 0x0f)
-#define IS_I82365	0x10
-#define IS_PD672X	0x20
-#define IS_PD6710	0x30
-#define IS_PD6729	0x40
-#define IS_VLSI		0x50
-#define IS_IBM		0x60
-#define IS_VG468	0x70
-#define IS_RICOH	0x80
+#define IS_I82365A	0x10
+#define IS_I82365B	0x11
+#define IS_I82365DF	0x12
+#define IS_VLSI		0x20
+#define IS_IBM		0x30
+#define IS_PD6710	0x40
+#define IS_PD672X	0x41
+#define IS_PD6729	0x42
+#define IS_VG468	0x50
+#define IS_VG469	0x51
+#define IS_RICOH	0x60
 
 /* Where we keep track of our sockets... */
 static int sockets = 0;
@@ -192,35 +197,32 @@ static void busy_loop(u_long len)
 
 static volatile u_long irq_hits;
 
-static void irq_count(int irq, struct pt_regs *regs)
+static void irq_count IRQ(int irq, void *dev, struct pt_regs *regs)
 {
     irq_hits++;
 }
 
 static u_long try_irq(u_short psock, int irq)
 {
-    u_char csc;
-    
-    if (request_irq(irq, irq_count, 0, "irq scan") != 0)
+    if (REQUEST_IRQ(irq, irq_count, 0, "irq scan") != 0)
 	return -1;
     irq_hits = 0;
-    busy_loop(2);
+    busy_loop(HZ/50);
     if (irq_hits) {
-	free_irq(irq);
+	FREE_IRQ(irq);
 	return -1;
     }
 
     /* Generate one interrupt */
-    csc = i365_get(psock, I365_CSCINT);
     i365_set(psock, I365_CSCINT, I365_CSC_DETECT | (irq << 4));
     i365_bclr(psock, I365_INTCTL, I365_INTR_ENA);
     i365_bset(psock, I365_GENCTL, I365_CTL_SW_IRQ);
 
-    busy_loop(2);
-    free_irq(irq);
+    busy_loop(HZ/50);
+    FREE_IRQ(irq);
 
     /* Turn off interrupts */
-    i365_set(psock, I365_CSCINT, csc);
+    i365_set(psock, I365_CSCINT, 0);
     
     return (irq_hits != 1);
 }
@@ -229,17 +231,25 @@ static u_long irq_scan(u_short lsock, u_long mask0)
 {
     u_short psock;
     u_long mask1, mask2;
-    int i, ni;
+    int i;
+
+#ifdef __alpha__
+#define PIC 0x4d0
+    /* Don't probe level-triggered interrupts -- reserved for PCI */
+    int level_mask = inb_p(PIC) | (inb_p(PIC+1) << 8);
+    if (level_mask)
+	mask0 &= ~level_mask;
+#endif
     
     psock = socket_table[lsock].psock;
-    mask1 = mask2 = ni = 0;
+    mask1 = mask2 = 0;
+    i365_set(psock, I365_CSCINT, 0);
     for (i = 0; i < 16; i++)
 	if ((mask0 & (1 << i)) && (try_irq(psock, i) == 0))
 	    mask1 |= (1 << i);
     for (i = 0; i < 16; i++)
 	if ((mask0 & (1 << i)) && (try_irq(psock, i) == 0)) {
 	    mask2 |= (1 << i);
-	    ni++;
 	}
     return (mask1 & mask2);
 }
@@ -267,49 +277,47 @@ static int identify(int sock)
     
     val = i365_get(sock, I365_IDENT);
     switch (val) {
-    case 0x82: case 0x83:
-	type = IS_I82365 | (val & 7); break;
+    case 0x82:
+	type = IS_I82365A; break;
+    case 0x83:
+	type = IS_I82365B; break;
     case 0x84:
 	type = IS_VLSI; break;
     case 0x88: case 0x89:
-	type = IS_IBM | (val & 7); break;
-    case 0x8b:
-	type = IS_VG468 | (val & 7);
-	i365_bclr(sock, VG468_MISC, VG468_MISC_VADEMREV);
-	break;
+	type = IS_IBM; break;
+    case 0x8b: case 0x8c:
+	/* Special: fall into Vadem test */
+	type = IS_VG468; break;
     default:
 	return -1;
     }
     
     /* Check for Vadem VG-468 chips */
-    if (CHIP(type) == IS_I82365) {
+    if ((type == IS_I82365B) || (type == IS_VG468)) {
 	outb_p(0x0e, I365_PORT(sock));
 	outb_p(0x37, I365_PORT(sock));
 	i365_bset(sock, VG468_MISC, VG468_MISC_VADEMREV);
 	val = i365_get(sock, I365_IDENT);
 	if (val & I365_IDENT_VADEM) {
-	    type = IS_VG468 | (val & 7);
+	    type = ((val & 7) == 4) ? IS_VG469 : IS_VG468;
 	    i365_bclr(sock, VG468_MISC, VG468_MISC_VADEMREV);
 	}
+	else if (type == IS_VG468)
+	    return -1;
     }
 
     /* Check for Ricoh chips */
     val = i365_get(sock, RF5C_CHIP_ID);
     if ((val == RF5C_CHIP_RF5C296) || (val == RF5C_CHIP_RF5C396))
-	type = IS_RICOH | ID(type);
+	type = IS_RICOH;
     
     /* Check for Cirrus CL-PD67xx chips */
     i365_set(sock, PD67_CHIP_INFO, 0);
     val = i365_get(sock, PD67_CHIP_INFO);
     if ((val & PD67_INFO_CHIP_ID) == PD67_INFO_CHIP_ID) {
 	val = i365_get(sock, PD67_CHIP_INFO);
-	if ((val & PD67_INFO_CHIP_ID) == 0) {
-	    if (val & PD67_INFO_SLOTS)
-		type = IS_PD672X;
-	    else
-		type = IS_PD6710;
-	    type |= 8 - ((val & PD67_INFO_REV) >> 2);
-	}
+	if ((val & PD67_INFO_CHIP_ID) == 0)
+	    type = (val & PD67_INFO_SLOTS) ? IS_PD672X : IS_PD6710;
     }
     return type;
 } /* identify */
@@ -318,45 +326,48 @@ static int identify(int sock)
 
 void add_pcic(int type, int ns)
 {
-    int i, j;
+    int i, j, irq;
     u_short base;
     u_long mask, scan;
     
-    if (sockets == ns) printk("\n  ");
-    switch (CHIP(type)) {
+    if (sockets == ns) printk("\n" KERN_INFO "  ");
+    switch (type) {
     case IS_VLSI:
 	printk("VLSI 82C146"); break;
-    case IS_I82365:
-	printk("Intel i82365sl"); break;
+    case IS_I82365A:
+	printk("Intel i82365sl A step"); break;
+    case IS_I82365B:
+	printk("Intel i82365sl B step"); break;
+    case IS_I82365DF:
+	printk("Intel i82365sl DF"); break;
+    case IS_PD6710:
+	printk("Cirrus PD6710"); break;
     case IS_PD672X:
 	printk("Cirrus PD672x");
 	if (has_dma) printk(", DMA");
 	if (has_led) printk(", LED");
 	if (has_ring) printk(", RING");
 	break;
-    case IS_PD6710:
-	printk("Cirrus PD6710"); break;
     case IS_PD6729:
 	printk("Cirrus PD6729 PCI"); break;
     case IS_IBM:
 	printk("IBM Clone"); break;
     case IS_VG468:
 	printk("Vadem VG-468"); break;
+    case IS_VG469:
+	printk("Vadem VG-469"); break;
     case IS_RICOH:
-	printk("Ricoh RF5C296/396");
-	if (has_led) printk(", LED");
-	break;
+	printk("Ricoh RF5C296/396"); break;
     }
     base = sockets-ns;
-    printk(" rev %d at %#x", ID(type),
-	   I365_PORT(socket_table[base].psock));
+    printk(" at %#x", I365_PORT(socket_table[base].psock));
     printk(", %d socket%s\n", ns, ((ns > 1) ? "s" : ""));
 
     /* Build interrupt mask; scan if possible */
-    printk("  irq mask (");
+    printk(KERN_INFO "  irq mask (");
     mask = i365_cap.irq_mask;
     if (irq_mask) mask &= irq_mask;
-    switch (CHIP(type)) {
+    switch (type) {
     case IS_PD6710:
     case IS_PD6729:
     case IS_PD672X:
@@ -368,8 +379,13 @@ void add_pcic(int type, int ns)
     default:
 	scan = irq_scan(base, mask);
 	if (scan) {
+	    u_long tmp;
 	    mask = scan;
 	    printk("scanned");
+	    /* Check for only two interrupts available */
+	    tmp = (scan & (scan-1));
+	    if (((tmp & (tmp-1)) == 0) && (poll_interval == 0))
+		poll_interval = HZ;
 	}
 	else
 	    printk("default");
@@ -382,28 +398,30 @@ void add_pcic(int type, int ns)
 	if (cs_irq == 0) {
 	    for (i = 15; i > 0; i--)
 		if ((mask & (1 << i)) &&
-		    (request_irq(i, i365_interrupt, 0, "i82365") == 0))
+		    (REQUEST_IRQ(i, i365_interrupt, 0, "i82365") == 0))
 		    break;
-	    cs_irq = i;
+	    irq = i;
 	}
 	else {
-	    if (request_irq(cs_irq, i365_interrupt, 0, "i82365") != 0)
-		cs_irq = 0;
+	    if (REQUEST_IRQ(cs_irq, i365_interrupt, 0, "i82365") != 0)
+		irq = cs_irq;
+	    else
+		irq = 0;
 	}
-	if (cs_irq == 0) poll_interval = 100;
+	if (irq == 0) poll_interval = 100;
     }
     else
-	cs_irq = 0;
+	irq = 0;
     
-    if (cs_irq)
-	printk("status change on irq %d\n", cs_irq);
+    if (irq)
+	printk("status change on irq %d\n", irq);
     else
 	printk("polled status, interval = %d\n", poll_interval);
     
     /* Turn on interrupts, but with all events masked */
     for (i = base; i < base+ns; i++) {
 	socket_table[i].cap.irq_mask = mask;
-	socket_table[i].cs_irq = cs_irq;
+	socket_table[i].cs_irq = irq;
 	if (cs_irq) {
 	    j = socket_table[i].psock;
 	    i365_set(j, I365_CSCINT, (cs_irq << 4));
@@ -414,19 +432,46 @@ void add_pcic(int type, int ns)
 
 } /* add_pcic */
 
+/*======================================================================
+
+    See if a card is present, powered up, in IO mode, and already
+    bound to a (non-PCMCIA) Linux driver.
+
+    We make an exception for cards that seem to be serial devices.
+    
+======================================================================*/
+
+static int is_active(int s)
+{
+    u_char stat;
+    u_short start, stop;
+    
+    stat = i365_get(s, I365_STATUS);
+    start = i365_get_pair(s, I365_IO(0)+I365_W_START);
+    stop = i365_get_pair(s, I365_IO(0)+I365_W_STOP);
+    if ((stat & I365_CS_DETECT) && (stat & I365_CS_POWERON) &&
+	(i365_get(s, I365_INTCTL) & I365_PC_IOCARD) &&
+	(i365_get(s, I365_ADDRWIN) & I365_ENA_IO(0)) &&
+	(check_region(start, stop-start+1) != 0) &&
+	((start & 0xfeef) != 0x02e8))
+	return 1;
+    else
+	return 0;
+}
+
 /*====================================================================*/
 
-int i365_init(void)
+static int i365_init(void)
 {
-    int i, j, k, ns;
+    int i, j, k, p, ns;
     int id, type, is_pci = 0;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     
-    printk("Intel PCIC probe: ");
+    printk(KERN_INFO "Intel PCIC probe: ");
     
 #ifdef CONFIG_PCI
     if (pcibios_present()) {
@@ -437,7 +482,7 @@ int i365_init(void)
 				0, &bus, &device_fn);
 	if (i == 0) {
 	    unsigned long addr;
-	    pcibios_read_config_dword(bus, device_fn, 0x10, &addr);
+	    pcibios_read_config_dword(bus, device_fn, 0x10, (void *)&addr);
 	    i365_base = addr & ~0x1;
 	    pcibios_write_config_dword(bus, device_fn, 0x4, 0x00ff00ff);
 	    is_pci = 1;
@@ -456,15 +501,15 @@ int i365_init(void)
     if (wakeup) {
 	i365_bclr(0, PD67_MISC_CTL_2, PD67_MC2_SUSPEND);
 	/* Pause at least 50 ms */
-	busy_loop(6);
+	busy_loop(HZ/20);
     }
     
     type = identify(0);
-    if (CHIP(type) == IS_VLSI) {
+    if (type == IS_VLSI) {
 	for (i = 0; i < 4; i++) {
 	    if (i == ignore) continue;
 	    j = ((i & 1) << 3) + (i << 1);
-	    if (identify(j) == type) {
+	    if ((identify(j) == type) && !is_active(j)) {
 		socket_table[sockets].psock = j;
 		socket_table[sockets].type = type;
 		sockets++;
@@ -473,36 +518,42 @@ int i365_init(void)
 	add_pcic(type, sockets);
     }
     else
-	for (i = 0; i < 8; i += 4) {
+	for (i = 0; i < (extra_sockets ? 8 : 4); i += 2) {
 	    id = identify(i);
 	    if (id < 0) continue;
 
-	    if (is_pci && (CHIP(id) == IS_PD672X))
-		type = ID(id) | IS_PD6729;
+	    if (is_pci && (id == IS_PD672X))
+		type = IS_PD6729;
 	    else
 		type = id;
 	    
 	    ns = 0;
-	    for (j = 0; j < 4; j++) {
+	    for (j = 0; j < 2; j++) {
 		/* Does the socket exist? */
-		if ((ignore == i+j) || (identify(i+j) != id))
+		if ((ignore == i+j) || (identify(i+j) < 0) ||
+		    is_active(i+j))
 		    continue;
 		/* Check for bad socket decode */
-		i365_set(i+j, I365_CSCINT, i+j);
-		for (k = i; k < i+j; k++) {
-		    if (i365_get(k, I365_CSCINT) != k) {
-			i365_set(k, I365_CSCINT, k);
+		i365_set(i+j, I365_MEM(0)+I365_W_OFF, i+j);
+		for (k = 0; k < sockets; k++) {
+		    p = socket_table[k].psock;
+		    if (i365_get(p, I365_MEM(0)+I365_W_OFF) != p) {
+			i365_set(p, I365_MEM(0)+I365_W_OFF, p);
 			break;
 		    }
 		}
-		if (k == i+j) {
+		if (i365_get(i+j, I365_MEM(0)+I365_W_OFF) != i+j)
+		    break;
+		if (k == sockets) {
 		    socket_table[sockets].psock = i+j;
 		    socket_table[sockets].type = type;
 		    sockets++;
 		    ns++;
 		}
 	    }
-	    add_pcic(type, ns);
+	    /* Some probes are more reliable than others... */
+	    if (ns != 0)
+		add_pcic(type, ns);
 	}
     
     if (sockets == 0) {
@@ -520,17 +571,20 @@ int i365_init(void)
     }
     
     if (register_ss_entry(sockets, &i365_service) != 0) {
-	printk("i82365: register_ss_entry() failed\n");
+	int this_irq, last_irq = 0;
+	printk(KERN_NOTICE "i82365: register_ss_entry() failed\n");
 	release_region(i365_base, 8);
-	for (i = 0; i < sockets; i++)
-	    if (((i & 3) == 0) && socket_table[i].cs_irq)
-		free_irq(socket_table[i].cs_irq);
+	for (i = 0; i < sockets; i++) {
+	    this_irq = socket_table[i].cs_irq;
+	    if (this_irq && (this_irq != last_irq))
+		FREE_IRQ(last_irq = this_irq);
+	}
 	return -ENODEV;
     }
 
     /* Finally, schedule a polling interrupt */
     if (poll_interval != 0) {
-    	poll_timer.expires = poll_interval;
+    	poll_timer.expires = RUN_AT(poll_interval);
 	add_timer(&poll_timer);
     }
     
@@ -542,38 +596,37 @@ int i365_init(void)
 
 static void i365_finish(void)
 {
-    int i;
+    int i, this_irq, last_irq = 0;
     unregister_ss_entry(&i365_service);
     if (poll_interval != 0)
 	del_timer(&poll_timer);
-    else
-	for (i = 0; i < sockets; i++) {
-	    i365_set(socket_table[i].psock, I365_CSCINT, 0);
-	    if (((i & 3) == 0) && socket_table[i].cs_irq)
-		free_irq(socket_table[i].cs_irq);
-	}
+    for (i = 0; i < sockets; i++) {
+	i365_set(socket_table[i].psock, I365_CSCINT, 0);
+	this_irq = socket_table[i].cs_irq;
+	if (this_irq && (this_irq != last_irq))
+	    FREE_IRQ(last_irq = this_irq);
+    }
     release_region(i365_base, 8);
 } /* i365_finish */
 
 /*====================================================================*/
 
-static void i365_interrupt(int irq, struct pt_regs *regs)
+static void i365_interrupt IRQ(int irq, void *dev, struct pt_regs *regs)
 {
     int i, csc;
     u_short psock;
     u_long events;
+    
+#ifdef PCMCIA_DEBUG
     static volatile int active = 0;
-
     if (active) {
-	printk("i82365: reentered interrupt handler!\n");
+	printk(KERN_NOTICE "i82365: reentered interrupt handler!\n");
 	return;
     }
     else
 	active = 1;
-
-#ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: i365_interrupt()\n");
+	printk(KERN_DEBUG "i82365: i365_interrupt()\n");
 #endif
     
     for (i = 0; i < sockets; i++) {
@@ -595,14 +648,14 @@ static void i365_interrupt(int irq, struct pt_regs *regs)
 
     /* Schedule next poll, if needed */
     if (poll_interval > 0) {
-	poll_timer.expires = poll_interval;
+	poll_timer.expires = RUN_AT(poll_interval);
 	add_timer(&poll_timer);
     }
-    active = 0;
     
 #ifdef PCMCIA_DEBUG
+    active = 0;
     if (pc_debug > 1)
-	printk("i82365: done\n");
+	printk(KERN_DEBUG "i82365: done\n");
 #endif
 
 } /* i365_interrupt */
@@ -613,14 +666,10 @@ static int i365_register_callback(u_short lsock, ss_callback_t *call)
 {
     if (call == NULL) {
 	socket_table[lsock].handler = NULL;
-#ifdef MODULE
 	MOD_DEC_USE_COUNT;
-#endif
     }
     else {
-#ifdef MODULE
 	MOD_INC_USE_COUNT;
-#endif
 	socket_table[lsock].handler = call->handler;
 	socket_table[lsock].info = call->info;
     }
@@ -648,7 +697,8 @@ static int i365_get_status(u_short lsock, u_int *value)
     *value |= (status & I365_CS_POWERON) ? SS_POWERON : 0;
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: GetStatus(%d) = %#2.2x\n", lsock, *value);
+	printk(KERN_DEBUG "i82365: GetStatus(%d) = %#2.2x\n",
+	       lsock, *value);
 #endif
     return 0;
 } /* i365_get_status */
@@ -673,13 +723,13 @@ static int i365_get_socket(u_short lsock, socket_state_t *state)
     state->flags |= (reg & I365_PWR_OUT) ? SS_OUTPUT_ENA : 0;
     vcc = reg & I365_VCC_MASK; vpp = reg & I365_VPP_MASK;
     state->Vcc = state->Vpp = 0;
-    switch (CHIP(socket_table[lsock].type)) {
+    switch (socket_table[lsock].type) {
     case IS_PD672X:
     case IS_PD6710:
     case IS_PD6729:
 	if (i365_get(psock, PD67_MISC_CTL_1) & PD67_MC1_VCC_3V) {
-	    if (reg & I365_VCC_5V) state->Vcc = 30;
-	    if (vpp == I365_VPP_5V) state->Vpp = 30;
+	    if (reg & I365_VCC_5V) state->Vcc = 33;
+	    if (vpp == I365_VPP_5V) state->Vpp = 33;
 	    if (vpp == I365_VPP_12V) state->Vpp = 120;
 	}
 	else {
@@ -688,19 +738,31 @@ static int i365_get_socket(u_short lsock, socket_state_t *state)
 	    if (vpp == I365_VPP_12V) state->Vpp = 120;
 	}
 	break;
-    case IS_VG468:
+    case IS_VG469:
+	if (i365_get(psock, VG469_VSELECT) & VG469_VSEL_VCC) {
+	    if (reg & I365_VCC_5V) state->Vcc = 33;
+	    if (vpp == I365_VPP_5V) state->Vpp = 33;
+	    if (vpp == I365_VPP_12V) state->Vpp = 120;
+	}
+	else {
+	    if (reg & I365_VCC_5V) state->Vcc = 50;
+	    if (vpp == I365_VPP_5V) state->Vpp = 50;
+	    if (vpp == I365_VPP_12V) state->Vpp = 120;
+	}
+	break;
+    case IS_I82365DF:
+	if (vcc == I365_VCC_3V) state->Vcc = 33;
+	if (vcc == I365_VCC_5V) state->Vcc = 50;
+	if (vpp == I365_VPP_5V) state->Vpp = 50;
+	if (vpp == I365_VPP_12V) state->Vpp = 120;
+	break;
+    default:
 	if (reg & I365_VCC_5V) {
 	    state->Vcc = 50;
 	    if (vpp == I365_VPP_5V) state->Vpp = 50;
 	    if (vpp == I365_VPP_12V) state->Vpp = 120;
 	}
 	break;
-    default:
-	if (vcc == I365_VCC_3V) state->Vcc = 33;
-	if (vcc == I365_VCC_5V) state->Vcc = 50;
-	if (vpp == I365_VPP_5V) state->Vpp = 50;
-	if (vpp == I365_VPP_12V) state->Vpp = 120;
-	break;
     }
     reg = i365_get(psock, I365_INTCTL);
     state->flags |= (reg & I365_PC_IOCARD) ? SS_IOCARD : 0;
@@ -713,7 +775,7 @@ static int i365_get_socket(u_short lsock, socket_state_t *state)
     state->io_irq = reg & I365_IRQ_MASK;
     
     /* DMA mode, speaker control */
-    switch (CHIP(socket_table[lsock].type)) {
+    switch (socket_table[lsock].type) {
     case IS_PD672X:
     case IS_PD6710:
     case IS_PD6729:
@@ -738,9 +800,10 @@ static int i365_get_socket(u_short lsock, socket_state_t *state)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: GetSocket(%d) = flags %#3.3lx, Vcc %d, Vpp %d"
-	       ", io_irq %d, csc_mask %#2.2lx\n", lsock, state->flags,
-	       state->Vcc, state->Vpp, state->io_irq, state->csc_mask);
+	printk(KERN_DEBUG "i82365: GetSocket(%d) = flags %#3.3lx, "
+	       "Vcc %d, Vpp %d, io_irq %d, csc_mask %#2.2lx\n", lsock,
+	       state->flags, state->Vcc, state->Vpp, state->io_irq,
+	       state->csc_mask);
 #endif
     return 0;
 } /* i365_get_socket */
@@ -749,16 +812,28 @@ static int i365_get_socket(u_short lsock, socket_state_t *state)
 
 static int i365_set_socket(u_short lsock, socket_state_t *state)
 {
+    int type = socket_table[lsock].type;
     u_short psock = socket_table[lsock].psock;
     u_char reg;
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: SetSocket(%d, flags %#3.3lx, Vcc %d, Vpp %d"
-	       ", io_irq %d, csc_mask %#2.2lx)\n", lsock, state->flags,
-	       state->Vcc, state->Vpp, state->io_irq, state->csc_mask);
+	printk(KERN_DEBUG "i82365: SetSocket(%d, flags %#3.3lx, "
+	       "Vcc %d, Vpp %d, io_irq %d, csc_mask %#2.2lx)\n",
+	       lsock, state->flags, state->Vcc, state->Vpp,
+	       state->io_irq, state->csc_mask);
 #endif
 
+    /* Did something put the controller to sleep? */
+    if ((type == IS_PD6710) || (type == IS_PD672X) ||
+	(type == IS_PD6729)) {
+	if (i365_get(0, PD67_MISC_CTL_2) & PD67_MC2_SUSPEND) {
+	    i365_bclr(0, PD67_MISC_CTL_2, PD67_MC2_SUSPEND);
+	    /* Pause at least 50 ms */
+	    busy_loop(HZ/20);
+	}
+    }
+    
     /* IO card, RESET flags, IO interrupt */
     reg = i365_get(psock, I365_INTCTL);
     reg &= ~(I365_PC_IOCARD | I365_PC_RESET | I365_IRQ_MASK);
@@ -772,11 +847,12 @@ static int i365_set_socket(u_short lsock, socket_state_t *state)
     if (state->flags & SS_PWR_AUTO) reg |= I365_PWR_AUTO;
     if (state->flags & SS_OUTPUT_ENA) reg |= I365_PWR_OUT;
     
-    switch (CHIP(socket_table[lsock].type)) {
+    switch (type) {
 
     case IS_PD6729:
     case IS_PD672X:
     case IS_PD6710:
+    case IS_VG469:
 	if (state->Vpp != 0) {
 	    if (state->Vpp == 120)
 		reg |= I365_VPP_12V;
@@ -785,19 +861,40 @@ static int i365_set_socket(u_short lsock, socket_state_t *state)
 	    else return -EINVAL;
 	}
 	if (state->Vcc != 0) {
-	    if (state->Vcc == 30) {
+	    if (state->Vcc == 33) {
 		reg |= I365_VCC_5V;
-		i365_bset(psock, PD67_MISC_CTL_1, PD67_MC1_VCC_3V);
+		if (type == IS_VG469)
+		    i365_bset(psock, VG469_VSELECT, VG469_VSEL_VCC);
+		else /* Cirrus */
+		    i365_bset(psock, PD67_MISC_CTL_1, PD67_MC1_VCC_3V);
 	    }
 	    else if (state->Vcc == 50) {
 		reg |= I365_VCC_5V;
-		i365_bclr(psock, PD67_MISC_CTL_1, PD67_MC1_VCC_3V);
+		if (type == IS_VG469)
+		    i365_bclr(psock, VG469_VSELECT, VG469_VSEL_VCC);
+		else /* Cirrus */
+		    i365_bclr(psock, PD67_MISC_CTL_1, PD67_MC1_VCC_3V);
 	    }
 	    else return -EINVAL;
 	}
 	break;
+
+    case IS_I82365DF:
+	switch (state->Vcc) {
+	case 0:		break;
+	case 33:   	reg |= I365_VCC_3V; break;
+	case 50:	reg |= I365_VCC_5V; break;
+	default:	return -EINVAL;
+	}
+	switch (state->Vpp) {
+	case 0:		break;
+	case 50:   	reg |= I365_VPP_5V; break;
+	case 120:	reg |= I365_VPP_12V; break;
+	default:	return -EINVAL;
+	}
+	break;
 	
-    case IS_VG468:
+    default:
 	switch (state->Vcc) {
 	case 0:		break;
 	case 50:	reg |= I365_VCC_5V; break;
@@ -811,28 +908,15 @@ static int i365_set_socket(u_short lsock, socket_state_t *state)
 	}
 	break;
 	
-    default:
-	switch (state->Vcc) {
-	case 0:		break;
-	case 30:   	reg |= I365_VCC_3V; break;
-	case 50:	reg |= I365_VCC_5V; break;
-	default:	return -EINVAL;
-	}
-	switch (state->Vpp) {
-	case 0:		break;
-	case 50:   	reg |= I365_VPP_5V; break;
-	case 120:	reg |= I365_VPP_12V; break;
-	default:	return -EINVAL;
-	}
-	break;
     }
     
     if (reg != i365_get(psock, I365_POWER))
 	i365_set(psock, I365_POWER, reg);
 
     /* Chipset-specific functions */
-    switch (CHIP(socket_table[lsock].type)) {
+    switch (type) {
     case IS_VG468:
+    case IS_VG469:
 	if (async_clock)
 	    i365_bset(psock, VG468_CTL, VG468_CTL_ASYNC);
 	else
@@ -923,7 +1007,7 @@ static int i365_get_io_map(u_short lsock, struct pcmcia_io_map *io)
     io->flags |= (ioctl & I365_IOCTL_IOCS16(map)) ? MAP_AUTOSZ : 0;
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: GetIOMap(%d, %d) = %#2.2x, %d ns, "
+	printk(KERN_DEBUG "i82365: GetIOMap(%d, %d) = %#2.2x, %d ns, "
 	       "%#4.4x-%#4.4x\n", lsock, map, io->flags,
 	       io->speed, io->start, io->stop);
 #endif
@@ -939,7 +1023,7 @@ static int i365_set_io_map(u_short lsock, struct pcmcia_io_map *io)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: SetIOMap(%d, %d, %#2.2x, %d ns, "
+	printk(KERN_DEBUG "i82365: SetIOMap(%d, %d, %#2.2x, %d ns, "
 	       "%#4.4x-%#4.4x)\n", lsock, io->map, io->flags,
 	       io->speed, io->start, io->stop);
 #endif
@@ -995,9 +1079,10 @@ static int i365_get_mem_map(u_short lsock, struct pcmcia_mem_map *mem)
     mem->card_start &= 0x3ffffff;
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: GetMemMap(%d, %d) = %#2.2x, %d ns, %#5.5lx-%#5.5lx,"
-	       " %#5.5lx\n", lsock, mem->map, mem->flags, mem->speed,
-	       mem->sys_start, mem->sys_stop, mem->card_start);
+	printk(KERN_DEBUG "i82365: GetMemMap(%d, %d) = %#2.2x, %d ns, "
+	       "%#5.5lx-%#5.5lx, %#5.5lx\n", lsock, mem->map,
+	       mem->flags, mem->speed, mem->sys_start, mem->sys_stop,
+	       mem->card_start);
 #endif
     return 0;
 } /* i365_get_mem_map */
@@ -1012,9 +1097,10 @@ static int i365_set_mem_map(u_short lsock, struct pcmcia_mem_map *mem)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("i82365: SetMemMap(%d, %d, %#2.2x, %d ns, %#5.5lx-%#5.5lx,"
-	       " %#5.5lx)\n", lsock, mem->map, mem->flags, mem->speed,
-	       mem->sys_start, mem->sys_stop, mem->card_start);
+	printk(KERN_DEBUG "i82365: SetMemMap(%d, %d, %#2.2x, %d ns, "
+	       "%#5.5lx-%#5.5lx, %#5.5lx)\n", lsock, mem->map,
+	       mem->flags, mem->speed, mem->sys_start, mem->sys_stop,
+	       mem->card_start);
 #endif
     map = mem->map;
     if ((map > 4) || (mem->card_start > 0x3ffffff) ||
@@ -1075,7 +1161,8 @@ static int i365_service(u_int lsock, u_int cmd, void *arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1) 
-	printk("i365_ioctl(%d, %d, 0x%p)\n", lsock, cmd, arg);
+	printk(KERN_DEBUG "i365_ioctl(%d, %d, 0x%p)\n",
+	       lsock, cmd, arg);
 #endif
 
     if (cmd < NFUNC)
@@ -1088,20 +1175,19 @@ static int i365_service(u_int lsock, u_int cmd, void *arg)
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_i82365(void)
 {
     servinfo_t serv;
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("i82365: Card Services release does not match!\n");
+	printk(KERN_NOTICE "i82365: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     return i365_init();
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
     i365_finish();
