@@ -6,11 +6,11 @@
     
 ======================================================================*/
 
-#include <linux/config.h>
+#include "config.h"
+#include "k_compat.h"
 
 #ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
+#define init_skeleton init_module
 #endif
 
 #include <linux/kernel.h>
@@ -38,7 +38,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-    "skeleton.c 1.17 1995/06/09 20:34:51 (David Hinds)\n";
+"skeleton.c 1.30 1996/05/28 15:25:32 (David Hinds)";
 #endif
 
 /*====================================================================*/
@@ -109,9 +109,19 @@ static dev_link_t *dev_list = NULL;
    specific information that also needs to be kept track of.  The
    'priv' pointer in a dev_link_t structure can be used to point to
    a device-specific private data structure, like this.
+
+   A driver needs to provide a dev_node_t structure for each device
+   on a card.  In some cases, there is only one device per card (for
+   example, ethernet cards, modems).  In other cases, there may be
+   many actual or logical devices (SCSI adapters, memory cards with
+   multiple partitions).  The dev_node_t structures need to be kept
+   in a linked list starting at the 'dev' field of a dev_link_t
+   structure.  We allocate them in the card's private data structure,
+   because they generally can't be allocated dynamically.
 */
    
 typedef struct local_info_t {
+    dev_node_t	*node;
 } local_info_t;
 
 /*====================================================================*/
@@ -142,7 +152,7 @@ static dev_link_t *skeleton_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("skeleton_attach()\n");
+	printk(KERN_DEBUG "skeleton_attach()\n");
 #endif
 
     /* Initialize the dev_link_t structure */
@@ -162,7 +172,8 @@ static dev_link_t *skeleton_attach(void)
     link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
     link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
     link->irq.IRQInfo2 = irq_mask;
-
+    link->irq.Handler = skeleton_interrupt;
+    
     /* General socket configuration */
     link->conf.Attributes = CONF_ENABLE_IRQ;
     link->conf.Vcc = 50;
@@ -212,7 +223,7 @@ static void skeleton_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("skeleton_detach(0x%p)\n", link);
+	printk(KERN_DEBUG "skeleton_detach(0x%p)\n", link);
 #endif
     
     /* Locate device structure */
@@ -228,8 +239,10 @@ static void skeleton_detach(dev_link_t *link)
        detach().
     */
     if (link->state & DEV_CONFIG) {
-	printk("skeleton_cs: detach postponed, '%s' still locked\n",
-	       link->dev_name);
+#ifdef PCMCIA_DEBUG
+	printk(KERN_DEBUG "skeleton_cs: detach postponed, '%s' "
+	       "still locked\n", link->dev->dev_name);
+#endif
 	link->state |= DEV_STALE_LINK;
 	return;
     }
@@ -272,7 +285,7 @@ static void skeleton_config(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("skeleton_config(0x%p)\n", link);
+	printk(KERN_DEBUG "skeleton_config(0x%p)\n", link);
 #endif
 
     /*
@@ -368,19 +381,14 @@ static void skeleton_config(dev_link_t *link)
 	    cs_error(MapMemPage, i);
 	    break;
 	    }
-
-	/*
-	   Finally, if everything else worked, we grab the interrupt
-	   vector.
-	*/
-	i = request_irq(link->irq.AssignedIRQ, &skeleton_interrupt,
-			0, "skeleton_cs");
-	if (i != CS_SUCCESS) {
-	    printk("skeleton_cs: unable to allocate irq %ld\n",
-		   link->irq.AssignedIRQ);
-	    break;
-	}
     } while (0);
+
+    /* At this point, the dev_node_t structure(s) should be
+       initialized and arranged in a linked list at link->dev. */
+    sprintf(dev->node.dev_name, "skel0");
+    dev->node.major = MAJOR;
+    dev->node.minor = 0;
+    link->dev = &dev->node;
     
     link->state &= ~DEV_CONFIG_PENDING;
     /* If any step failed, release any partially configured state */
@@ -389,7 +397,7 @@ static void skeleton_config(dev_link_t *link)
 	return;
     }
 
-    printk("skeleton device loaded\n");
+    printk(KERN_DEBUG "skeleton device loaded\n");
 } /* skeleton_config */
 
 /*======================================================================
@@ -407,7 +415,7 @@ static void skeleton_release(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("skeleton_release(0x%p)\n", link);
+	printk(KERN_DEBUG "skeleton_release(0x%p)\n", link);
 #endif
 
     /*
@@ -415,20 +423,21 @@ static void skeleton_release(u_long arg)
        is actually closed.
     */
     if (link->open) {
-	printk("skeleton_cs: release postponed, '%s' still open\n",
-	       link->dev_name);
+	if (pc_debug)
+	    printk(KERN_DEBUG "skeleton_cs: release postponed, '%s' "
+		   "still open\n", link->dev->dev_name);
 	link->state |= DEV_STALE_CONFIG;
 	return;
     }
 
+    /* Unlink the device chain */
+    link->dev = NULL;
+    
     /* Don't bother checking to see if these succeed or not */
     CardServices(ReleaseWindow, link->win);
     CardServices(ReleaseConfiguration, link->handle);
     CardServices(ReleaseIO, link->handle, &link->io);
     CardServices(ReleaseIRQ, link->handle, &link->irq);
-    if (link->irq.AssignedIRQ != 0) {
-	free_irq(link->irq.AssignedIRQ);
-    }
     link->state &= ~DEV_CONFIG;
     
     if (link->state & DEV_STALE_LINK)
@@ -457,21 +466,21 @@ static int skeleton_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("skeleton_event()\n");
+	printk(KERN_DEBUG "skeleton_event()\n");
 #endif
     
     switch (event) {
 #ifdef PCMCIA_DEBUG
     case CS_EVENT_REGISTRATION_COMPLETE:
 	if (pc_debug)
-	    printk("skeleton_cs: registration complete\n");
+	    printk(KERN_DEBUG "skeleton_cs: registration complete\n");
 	break;
 #endif
     case CS_EVENT_CARD_REMOVAL:
 	link->state &= ~DEV_PRESENT;
 	if (link->state & DEV_CONFIG) {
 	    ((local_info_t *)link->priv)->block = 1;
-	    link->release.expires = 5;
+	    link->release.expires = RUN_AT(HZ/20);
 	    add_timer(&link->release);
 	}
 	break;
@@ -501,33 +510,35 @@ static int skeleton_event(event_t event, int priority,
 
 void skeleton_interrupt(int reg)
 {
-    printk("skeleton_cs: interrupt\n");
+    printk(KERN_DEBUG "skeleton_cs: interrupt\n");
 } /* skeleton_interrupt */
 
 /*====================================================================*/
 
-#ifdef MODULE
-char kernel_version[] = UTS_RELEASE;
-
-int init_module(void)
+int init_skeleton(void)
 {
     servinfo_t serv;
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk(version);
+	printk(KERN_INFO "%s\n", version);
 #endif
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("skeleton: Card Services release does not match!\n");
+	printk(KERN_NOTICE "skeleton: Card Services release "
+	       "does not match!\n");
 	return -1;
     }
     register_pcmcia_driver(&dev_info, &skeleton_attach, &skeleton_detach);
     return 0;
 }
 
+#ifdef MODULE
 void cleanup_module(void)
 {
-    printk("skeleton_cs: unloading\n");
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk(KERN_DEBUG "skeleton_cs: unloading\n");
+#endif
     unregister_pcmcia_driver(&dev_info);
     while (dev_list != NULL) {
 	if (dev_list->state & DEV_CONFIG)
